#include <omp.h>

typedef double (*fprototype)(double);
int maxrecur = 0;

#pragma omp declare target
inline double f1(double);
int maxrecur;
double integrate( fprototype f, double a, double b, double tau, int recurstep );
void measure( char *prefix );
#pragma omp end declare target

inline double f1( double x )
{
	return 4.0/(x*x+1.0);
}

#define TAU 1E-7
double integrate( fprototype f, double a, double b, double tau, int recurstep )
{
	double h = b-a;
	double approx = h * f(b); // a simple rectangle
	double approx2 = h/2.0 * ( f(a) + f(b) ); // trapez
	double error = fabs(approx - approx2); // difference
	if( error > tau )
	{
		double p = (b+a)/2.0;
		double r1 = 0;
		double r2 = 0;
		if( recurstep < maxrecur )
		{
			#pragma omp task shared(r1)
			{
				r1 = integrate( f, a, p, tau/2.0, recurstep+1 );
			}
			#pragma omp task shared(r2)
			{
				r2 = integrate( f, p, b, tau/2.0, recurstep+1 );
			}
			#pragma omp taskwait
			return r1+r2;
		}
		else
		{
			return 	integrate( f, a, p, tau/2.0, recurstep+1 ) + 
				integrate( f, p, b, tau/2.0, recurstep+1 );
		}
	}
	return approx2;
}

void measure( char *prefix )
{
	#pragma omp master
	{
		#pragma omp task
		{
			double t = omp_get_wtime();
			double pi;
			pi = integrate( f1, 0, 1, TAU, 0 );
			double runtime = omp_get_wtime()-t;
			printf("%s: %f, TIME: %f w. %d threads and maxrecur: %d\n", prefix, pi, runtime, omp_get_num_threads(), maxrecur);
		}
		#pragma omp taskwait
	}
}

int main( int argc, char **argv )
{
	#pragma omp target 
	{
		maxrecur = 64;
		#pragma omp parallel num_threads(240)
		{
			measure("MIC");
		}
	}
	return 0;
}
